import os
from collections import defaultdict
from pathlib import Path

from supervisely import ProjectMeta, logger
from supervisely.annotation.obj_class import ObjClass
from supervisely.convert.volume.nii import nii_volume_helper as helper
from supervisely.convert.volume.nii.nii_volume_converter import NiiConverter
from supervisely.convert.volume.volume_converter import VolumeConverter
from supervisely.geometry.mask_3d import Mask3D
from supervisely.io.fs import get_file_ext, get_file_name
from supervisely.volume.volume import is_nifti_file
from supervisely.volume_annotation.volume_annotation import VolumeAnnotation
from supervisely.volume_annotation.volume_object import VolumeObject


class NiiPlaneStructuredConverter(NiiConverter, VolumeConverter):
    """Convert NIfTI 3D volume file to Supervisely format.
    The NIfTI file should be structured as follows:
    - <prefix>_anatomic_<idx>.nii (or .nii.gz)
    - <prefix>_inference_<idx>.nii (or .nii.gz)
        where   <prefix> is one of the following: cor, sag, axl
                <idx> is the index of the volume (to match volumes with annotations)

    Supports .nii and .nii.gz files.

    Example:
        📂 .
        ├── 🩻 axl_anatomic_1.nii
        ├── 🩻 axl_inference_1.nii class 1 (may contain multiple instances of the same class)
        ├── 🩻 cor_anatomic_1.nii
        ├── 🩻 cor_inference_1.nii class 1
        ├── 🩻 sag_anatomic_1.nii
        ├── 🩻 sag_inference_1.nii class 1
        ├── 🩻 sag_inference_2.nii class 2
        └── 🩻 sag_inference_3.nii class 3

    Additionally, if a TXT file with class color map is present, it will be used to
    create the classes with names and colors corresponding to the pixel values in the NIfTI files.
    The TXT file should be structured as follows:

    ```txt
    1 Femur 255 0 0
    2 Femoral cartilage 0 255 0
    3 Tibia 0 0 255
    4 Tibia cartilage 255 255 0
    5 Patella 0 255 255
    6 Patellar cartilage 255 0 255
    7 Miniscus 175 175 175
    ```
    where   1, 2, ... are the pixel values in the NIfTI files
            Femur, Femoral cartilage, ... are the names of the classes
            255, 0, 0, ... are the RGB colors of the classes
    The class name will be used to create the corresponding ObjClass in Supervisely.
    """

    class Item(VolumeConverter.BaseItem):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self._is_semantic = False

        @property
        def is_semantic(self) -> bool:
            return self._is_semantic

        @is_semantic.setter
        def is_semantic(self, value: bool):
            self._is_semantic = value

    def validate_format(self) -> bool:
        # create Items
        converted_dir_name = "converted"

        volumes_dict = defaultdict(list)
        ann_dict = defaultdict(list)
        cls_color_map = None

        for root, _, files in os.walk(self._input_data):
            if converted_dir_name in root:
                continue
            for file in files:
                path = os.path.join(root, file)
                if is_nifti_file(path):
                    full_name = get_file_name(path)
                    if full_name.endswith(".nii"):
                        full_name = get_file_name(full_name)
                    prefix = full_name.split("_")[0]
                    if prefix not in helper.PlanePrefix.values():
                        continue
                    name = full_name.split("_")[1]
                    if name in helper.LABEL_NAME or name[:-1] in helper.LABEL_NAME:
                        ann_dict[prefix].append(path)
                    else:
                        volumes_dict[prefix].append(path)
                ext = get_file_ext(path)
                if ext == ".txt":
                    cls_color_map = helper.read_cls_color_map(path)
                    if cls_color_map is None:
                        logger.warning(f"Failed to read class color map from {path}.")

        self._items = []
        for prefix, paths in volumes_dict.items():
            if len(paths) == 1:
                item = self.Item(item_path=paths[0])
                item.ann_data = ann_dict.get(prefix, [])
                item.is_semantic = len(item.ann_data) == 1
                if cls_color_map is not None:
                    item.custom_data["cls_color_map"] = cls_color_map
                self._items.append(item)
            elif len(paths) > 1:
                logger.info(
                    f"Found {len(paths)} volumes with prefix {prefix}. Will try to match them by directories."
                )
                for path in paths:
                    item = self.Item(item_path=path)
                    possible_ann_paths = []
                    for ann_path in ann_dict.get(prefix):
                        if Path(ann_path).parent == Path(path).parent:
                            possible_ann_paths.append(ann_path)
                    item.ann_data = possible_ann_paths
                    item.is_semantic = len(possible_ann_paths) == 1
                    if cls_color_map is not None:
                        item.custom_data["cls_color_map"] = cls_color_map
                    self._items.append(item)
        self._meta = ProjectMeta()
        return self.items_count > 0

    def to_supervisely(
        self,
        item: VolumeConverter.Item,
        meta: ProjectMeta = None,
        renamed_classes: dict = None,
        renamed_tags: dict = None,
    ) -> VolumeAnnotation:
        """Convert to Supervisely format."""

        try:
            objs = []
            spatial_figures = []
            for idx, ann_path in enumerate(item.ann_data, start=1):
                for mask, pixel_id in helper.get_annotation_from_nii(ann_path):
                    class_id = pixel_id if item.is_semantic else idx
                    class_name = f"Segment_{class_id}"
                    color = None
                    if item.custom_data.get("cls_color_map") is not None:
                        class_info = item.custom_data["cls_color_map"].get(class_id)
                        if class_info is not None:
                            class_name, color = class_info
                    class_name = renamed_classes.get(class_name, class_name)
                    obj_class = meta.get_obj_class(class_name)
                    if obj_class is None:
                        obj_class = ObjClass(class_name, Mask3D, color)
                        meta = meta.add_obj_class(obj_class)
                        self._meta_changed = True
                        self._meta = meta
                    obj = VolumeObject(obj_class, mask_3d=mask)
                    spatial_figures.append(obj.figure)
                    objs.append(obj)
            return VolumeAnnotation(item.volume_meta, objects=objs, spatial_figures=spatial_figures)
        except Exception as e:
            logger.warning(f"Failed to convert {item.path} to Supervisely format: {e}")
            return item.create_empty_annotation()



import datetime
from typing import List

from supervisely.nn.benchmark.object_detection.base_vis_metric import DetectionVisMetric
from supervisely.nn.benchmark.visualization.widgets import MarkdownWidget, TableWidget


class Overview(DetectionVisMetric):

    def get_header(self, user_login: str) -> MarkdownWidget:
        current_date = datetime.datetime.now().strftime("%d %B %Y, %H:%M")
        header_text = self.vis_texts.markdown_header.format(
            self.eval_result.name, user_login, current_date
        )
        header = MarkdownWidget("markdown_header", "Header", text=header_text)
        return header

    @property
    def md(self) -> List[MarkdownWidget]:
        url = self.eval_result.inference_info.get("checkpoint_url")
        link_text = self.eval_result.inference_info.get("custom_checkpoint_path")
        if link_text is None:
            link_text = url or ""
        link_text = link_text.replace("_", "\_")

        model_name = self.eval_result.inference_info.get("model_name") or "Custom"
        checkpoint_name = self.eval_result.checkpoint_name

        # Note about validation dataset
        classes_str, note_about_images, starter_app_info = self._get_overview_info()

        # link to scroll to the optimal confidence section
        opt_conf_url = self.vis_texts.docs_url + "#f1-optimal-confidence-threshold"
        average_url = self.vis_texts.docs_url + "#averaging-iou-thresholds"

        iou_threshold = self.eval_result.mp.iou_threshold
        if self.eval_result.different_iou_thresholds_per_class:
            iou_threshold = "Different IoU thresholds for each class (see the table below)"

        conf_text = (
            f"- **Optimal confidence threshold**: "
            f"{round(self.eval_result.mp.f1_optimal_conf, 4)} (calculated automatically), "
            f"<a href='{opt_conf_url}' target='_blank'>learn more</a>."
        )
        custom_conf_thrs = self.eval_result.mp.custom_conf_threshold
        if custom_conf_thrs is not None:
            conf_text += f"\n- **Custom confidence threshold**: {custom_conf_thrs}"

        formats = [
            model_name.replace("_", "\_"),
            checkpoint_name.replace("_", "\_"),
            self.eval_result.inference_info.get("architecture"),
            self.eval_result.inference_info.get("task_type"),
            self.eval_result.inference_info.get("runtime"),
            url,
            link_text,
            self.eval_result.gt_project_info.id,
            self.eval_result.gt_project_info.name,
            classes_str,
            note_about_images,
            starter_app_info,
            iou_threshold,
            conf_text,
            self.eval_result.mp.average_across_iou_thresholds,
            average_url,
            self.vis_texts.docs_url,
        ]

        md = MarkdownWidget(
            "markdown_overview",
            "Overview",
            text=self.vis_texts.markdown_overview.format(*formats),
        )
        md.is_info_block = True
        md.width_fit_content = True
        return md

    def _get_overview_info(self):
        classes_cnt = len(self.eval_result.classes_whitelist)
        classes_str = "classes" if classes_cnt > 1 else "class"
        classes_str = f"{classes_cnt} {classes_str}"

        evaluator_session, train_session, images_str = None, None, ""
        gt_project_id = self.eval_result.gt_project_info.id
        gt_dataset_ids = self.eval_result.gt_dataset_ids
        gt_images_cnt = self.eval_result.val_images_cnt
        train_info = self.eval_result.train_info
        evaluator_app_info = self.eval_result.evaluator_app_info
        total_imgs_cnt = self.eval_result.gt_project_info.items_count
        if gt_images_cnt is not None:
            val_imgs_cnt = gt_images_cnt
        elif gt_dataset_ids is not None:
            datasets = self.eval_result.gt_dataset_infos
            val_imgs_cnt = sum(ds.items_count for ds in datasets)
        else:
            val_imgs_cnt = self.eval_result.pred_project_info.items_count

        if train_info:
            train_task_id = train_info.get("app_session_id")
            if train_task_id:
                app_id = self.eval_result.task_info["meta"]["app"]["id"]
                train_session = f'- **Training dashboard**:  <a href="/apps/{app_id}/sessions/{train_task_id}" target="_blank">open</a>'

            train_imgs_cnt = train_info.get("images_count")
            images_str = f", {train_imgs_cnt} images in train, {val_imgs_cnt} images in validation"

        if gt_images_cnt is not None:
            images_str += (
                f", total {total_imgs_cnt} images. Evaluated using subset - {gt_images_cnt} images"
            )
        elif gt_dataset_ids is not None:
            links = [
                f'<a href="/projects/{gt_project_id}/datasets/{ds.id}" target="_blank">{ds.name}</a>'
                for ds in datasets
            ]
            images_str += f", total {total_imgs_cnt} images. Evaluated on the dataset{'s' if len(links) > 1 else ''}: {', '.join(links)}"
        else:
            images_str += f", total {total_imgs_cnt} images. Evaluated on the whole project ({val_imgs_cnt} images)"

        if evaluator_app_info:
            evaluator_task_id = evaluator_app_info.get("id")
            evaluator_app_id = evaluator_app_info.get("meta", {}).get("app", {}).get("id")
            evaluator_app_name = evaluator_app_info.get("meta", {}).get("app", {}).get("name")
            if evaluator_task_id and evaluator_app_id and evaluator_app_name:
                evaluator_session = f'- **Evaluator app session**:  <a href="/apps/{evaluator_app_id}/sessions/{evaluator_task_id}" target="_blank">open</a>'

        starter_app_info = train_session or evaluator_session or ""

        return classes_str, images_str, starter_app_info

    @property
    def iou_per_class_md(self) -> List[MarkdownWidget]:
        if not self.eval_result.different_iou_thresholds_per_class:
            return None

        return MarkdownWidget(
            "markdown_iou_per_class",
            "Different IoU thresholds for each class",
            text=self.vis_texts.markdown_iou_per_class,
        )

    @property
    def iou_per_class_table(self) -> TableWidget:
        if not self.eval_result.different_iou_thresholds_per_class:
            return None

        content = []
        for name, thr in self.eval_result.mp.iou_threshold_per_class.items():
            row = [name, round(thr, 2)]
            dct = {"row": row, "id": name, "items": row}
            content.append(dct)

        data = {
            "columns": ["Class name", "IoU threshold"],
            "columnsOptions": [{"disableSort": True}, {}],
            "content": content,
        }
        return TableWidget(
            name="table_iou_per_class",
            data=data,
            fix_columns=1,
            width="60%",
            main_column="Class name",
        )



import json
import os
from os.path import join as pjoin
from typing import Callable, List, Optional

import numpy as np

from supervisely import Bitmap
from supervisely._utils import batched
from supervisely.project.project import Dataset, OpenMode, Project


def sly2coco(
    sly_project_path: str,
    is_dt_dataset: bool,
    accepted_shapes: list = None,
    conf_threshold: float = None,
    progress: Optional[Callable] = None,
    classes_whitelist: Optional[List[str]] = None,
):
    from pycocotools import mask as maskUtils  # pylint: disable=import-error

    project = Project(sly_project_path, mode=OpenMode.READ)

    # Categories
    meta_path = pjoin(sly_project_path, "meta.json")
    with open(meta_path, "r") as f:
        meta = json.load(f)
    classes_sorted = sorted(meta["classes"], key=lambda x: x["title"])
    if accepted_shapes is not None:
        accepted_shapes = set(accepted_shapes)
        accepted_shapes.add("any")

    cat_id_map = {}
    i = 1
    for obj_cls in classes_sorted:
        if accepted_shapes is not None and obj_cls["shape"] not in accepted_shapes:
            continue
        if classes_whitelist:
            if obj_cls["title"] not in classes_whitelist:
                continue
        cat_id_map[obj_cls["title"]] = i
        i += 1

    categories = [{"id": id, "name": cat} for cat, id in cat_id_map.items()]

    # Images + Annotations
    images = []
    annotations = []
    annotation_id = 1
    total = project.total_items
    with progress(message="Evaluation: Converting to COCO format", total=total) as pbar:
        img_id = 1
        for dataset in sorted(project.datasets, key=lambda x: x.name):
            dataset: Dataset
            ann_path = dataset.ann_dir
            imginfo_path = dataset.img_info_dir
            ann_files = sorted(os.listdir(ann_path))
            for batch in batched(ann_files, 30):
                for ann_file in batch:
                    img_name = os.path.splitext(ann_file)[0]
                    with open(os.path.join(ann_path, ann_file), "r") as f:
                        ann = json.load(f)
                    with open(os.path.join(imginfo_path, ann_file), "r") as f:
                        img_info = json.load(f)
                    img_w = ann["size"]["width"]
                    img_h = ann["size"]["height"]
                    img = {
                        "id": img_id,
                        "file_name": img_name,
                        "width": img_w,
                        "height": img_h,
                        "sly_id": img_info["id"],
                        "dataset": dataset.name,
                    }
                    images.append(img)
                    for label in ann["objects"]:
                        geometry_type = label["geometryType"]
                        if accepted_shapes is not None and geometry_type not in accepted_shapes:
                            continue
                        class_name = label["classTitle"]
                        if classes_whitelist:
                            if class_name not in classes_whitelist:
                                continue
                        category_id = cat_id_map[class_name]
                        sly_id = label["id"]
                        if geometry_type == "rectangle":
                            ((left, top), (right, bottom)) = label["points"]["exterior"]
                            width = right - left + 1
                            height = bottom - top + 1
                            annotation = {
                                "id": annotation_id,
                                "image_id": img_id,
                                "category_id": category_id,
                                "bbox": [left, top, width, height],
                                "area": float(width * height),
                                "iscrowd": 0,
                                "sly_id": sly_id,
                            }
                        elif geometry_type in ["polygon", "bitmap"]:
                            if geometry_type == "bitmap":
                                bitmap = Bitmap.from_json(label)
                                mask_np = _uncrop_bitmap(bitmap, img_w, img_h)
                                segmentation = maskUtils.encode(np.asfortranarray(mask_np))
                            else:
                                polygon = label["points"]["exterior"]
                                polygon = [[coord for sublist in polygon for coord in sublist]]
                                rles = maskUtils.frPyObjects(polygon, img_h, img_w)
                                segmentation = maskUtils.merge(rles)
                            segmentation["counts"] = segmentation["counts"].decode()
                            annotation = {
                                "id": annotation_id,
                                "image_id": img_id,
                                "category_id": category_id,
                                "segmentation": segmentation,
                                "iscrowd": 0,
                                "sly_id": sly_id,
                            }
                            if not is_dt_dataset:
                                area = int(maskUtils.area(segmentation))
                                bbox = maskUtils.toBbox(segmentation)
                                bbox = [int(coord) for coord in bbox]
                                annotation["area"] = area
                                annotation["bbox"] = bbox
                        else:
                            raise NotImplementedError(
                                f"Geometry type '{geometry_type}' is not implemented."
                            )
                        # Extract confidence score from the tag
                        if is_dt_dataset:
                            conf = _extract_confidence(label)
                            annotation["score"] = conf
                            if conf_threshold is not None and conf < conf_threshold:
                                continue
                        annotations.append(annotation)
                        annotation_id += 1
                    img_id += 1

                pbar.update(len(batch))

    coco_dataset = {"images": images, "annotations": annotations, "categories": categories}
    return coco_dataset


def _extract_confidence(label: dict):
    conf_tag = [tag for tag in label["tags"] if tag["name"] in ["confidence", "confidence-model"]]
    assert len(conf_tag) == 1, f"'confidence' tag is not found."
    return float(conf_tag[0]["value"])


def _uncrop_bitmap(bitmap: Bitmap, image_width, image_height):
    data = bitmap.data
    h, w = data.shape
    o = bitmap.origin
    b = np.zeros((image_height, image_width), dtype=data.dtype)
    b[o.row : o.row + h, o.col : o.col + w] = data
    return b



import os
from typing import Generator

import nrrd
import numpy as np

from supervisely.collection.str_enum import StrEnum
from supervisely.geometry.mask_3d import Mask3D
from supervisely.io.fs import ensure_base_path, get_file_ext, get_file_name
from supervisely.sly_logger import logger
from supervisely.volume.volume import convert_3d_nifti_to_nrrd

VOLUME_NAME = "anatomic"
LABEL_NAME = ["inference", "label", "annotation", "mask", "segmentation"]


class PlanePrefix(str, StrEnum):
    """Prefix for plane names."""

    CORONAL = "cor"
    SAGITTAL = "sag"
    AXIAL = "axl"


def read_cls_color_map(path: str) -> dict:
    """Read class color map from TXT file.

    ```txt
    1 Femur 255 0 0
    2 Femoral cartilage 0 255 0
    3 Tibia 0 0 255
    4 Tibia cartilage 255 255 0
    5 Patella 0 255 255
    6 Patellar cartilage 255 0 255
    7 Miniscus 175 175 175
    ```
    """

    cls_color_map = {}
    if not os.path.exists(path):
        return None
    try:
        with open(path, "r") as file:
            for line in file:
                parts = line.strip().split()
                color = None
                try:
                    color = list(map(int, parts[-3:]))
                except:
                    pass
                cls_id = int(parts[0])
                if not color:
                    cls_name = " ".join(parts[1:])
                else:
                    cls_name = " ".join(parts[1:-3])
                if cls_id in cls_color_map:
                    logger.warning(f"Duplicate class ID {cls_id} found in color map.")
                if cls_name in cls_color_map:
                    logger.warning(f"Duplicate class name {cls_name} found in color map.")
                if len(color) != 3:
                    logger.warning(f"Invalid color format for class {cls_name}. Expected 3 values.")
                if any(c < 0 or c > 255 for c in color):
                    logger.warning(
                        f"Invalid color value for class {cls_name}. Expected values between 0 and 255."
                    )
                cls_color_map[cls_id] = (cls_name, color)
    except Exception as e:
        logger.warning(f"Failed to read class color map from {path}: {e}")
        return None
    return cls_color_map


def nifti_to_nrrd(nii_file_path: str, converted_dir: str) -> str:
    """Convert NIfTI 3D volume file to NRRD 3D volume file."""

    output_name = get_file_name(nii_file_path)
    if get_file_ext(output_name) == ".nii":
        output_name = get_file_name(output_name)

    data, header = convert_3d_nifti_to_nrrd(nii_file_path)

    nrrd_file_path = os.path.join(converted_dir, f"{output_name}.nrrd")
    ensure_base_path(nrrd_file_path)

    nrrd.write(nrrd_file_path, data, header)
    return nrrd_file_path


def get_annotation_from_nii(path: str) -> Generator[Mask3D, None, None]:
    """Get annotation from NIfTI 3D volume file."""

    data, _ = convert_3d_nifti_to_nrrd(path)
    unique_classes = np.unique(data)

    for class_id in unique_classes:
        if class_id == 0:
            continue
        mask = Mask3D(data == class_id)
        yield mask, class_id



from dataclasses import dataclass
from typing import List

from supervisely.nn.artifacts.artifacts import TrainInfo
from supervisely.nn.experiments import ExperimentInfo


class ModelSource:
    PRETRAINED = "Pretrained models"
    CUSTOM = "Custom models"


class RuntimeType:
    PYTORCH = "PyTorch"
    ONNXRUNTIME = "ONNXRuntime"
    TENSORRT = "TensorRT"


class ModelPrecision:
    FP32 = "FP32"
    FP16 = "FP16"
    INT8 = "INT8"


@dataclass
class CheckpointInfo:
    """
    :param checkpoint_name: A name for model weights, e.g., "YOLOv8n COCO (best)".
    :param model_name: The name of a model for which the weights are applicable, e.g., "YOLOv8n".
    :param architecture: Collection for a set of models, e.g., "YOLOv8".
    :param checkpoint_url: URL to download the model weights.
    :param custom_checkpoint_path: Path in Team Files to the weights.
    :param model_source: Source of the model, either "Pretrained models" or "Custom models".
    """

    checkpoint_name: str = None
    model_name: str = None
    architecture: str = None
    checkpoint_url: str = None
    custom_checkpoint_path: str = None
    model_source: str = None


@dataclass
class DeployInfo:
    checkpoint_name: str
    model_name: str
    architecture: str
    checkpoint_url: str
    custom_checkpoint_path: str
    model_source: str
    task_type: str
    device: str
    runtime: str
    model_precision: str
    hardware: str
    deploy_params: dict


def _get_model_name(model_info: dict):
    name = model_info.get("model_name")
    if not name:
        name = model_info.get("meta", {}).get("model_name")
    if not name:
        raise ValueError("Model name not found not in model_info nor in meta.")
    return name



# coding: utf-8

from supervisely.api.module_api import ApiField
from supervisely.api.pointcloud.pointcloud_tag_api import PointcloudObjectTagApi
from typing import List


class PointcloudEpisodeObjectTagApi(PointcloudObjectTagApi):
    _entity_id_field = ApiField.OBJECT_ID
    _method_bulk_add = "annotation-objects.tags.bulk.add"

    def update_frame_range(self, id: int, frame_range: List[int]) -> None:
        """Update tag frame range for annotation object in point cloud episode.

        :param id: unique ID of the tag specifically created for the object
        :type id: int
        :param frame_range: range of possible frames, it must always have strictly 2 values
        :type frame_range: List[int]
        """
        request_body = {
            ApiField.ID: id,
            ApiField.FRAME_RANGE: frame_range,
        }
        self._api.post("annotation-objects.tags.update", request_body)



from supervisely.convert.base_converter import AvailableVideoConverters, BaseConverter


class DavisConverter(BaseConverter):

    def __str__(self):
        return AvailableVideoConverters.DAVIS

    def get_meta(self):
        return super().get_meta()

    def get_items(self):
        return super().get_items()

    def to_supervisely(self, image_path: str, ann_path: str):
        raise NotImplementedError()



# Volume
from supervisely.convert.volume.sly.sly_volume_converter import SLYVolumeConverter
from supervisely.convert.volume.dicom.dicom_converter import DICOMConverter
from supervisely.convert.volume.nii.nii_volume_converter import NiiConverter
from supervisely.convert.volume.nii.nii_planes_volume_converter import (
    NiiPlaneStructuredConverter,
)



import argparse
import os
from contextlib import contextmanager
from typing import Dict, List, Union

import cv2
import numpy as np

from supervisely import (
    Annotation,
    Frame,
    FrameCollection,
    Label,
    Rectangle,
    VideoAnnotation,
    VideoFigure,
    VideoObject,
    VideoObjectCollection,
)
from supervisely.sly_logger import logger


class BaseDetection:
    """
    This class represents a bounding box detection in a single image.

    Parameters
    ----------
    tlwh : array_like
        Bounding box in format `(x, y, w, h)`.
    confidence : float
        Detector confidence score.
    feature : array_like | NoneType
        A feature vector that describes the object contained in this image.
    sly_label : Label | NoneType
        A Supervisely Label object

    Attributes
    ----------
    tlwh : ndarray
        Bounding box in format `(top left x, top left y, width, height)`.
    confidence : ndarray
        Detector confidence score.
    feature : ndarray | NoneType
        A feature vector that describes the object contained in this image.
    sly_label : Label | NoneType
        A Supervisely Label object

    """

    def __init__(self, tlwh, confidence: float, feature=None, sly_label: Label = None):
        self.tlwh = np.asarray(tlwh, dtype=float)
        self.confidence = float(confidence)
        self.feature = np.asarray(feature, dtype=np.float32)
        self._sly_label = sly_label

    def __iter__(self):
        return iter([*self.tlwh, self.confidence, self.feature])

    def tlbr(self):
        """Convert bounding box to format `(min x, min y, max x, max y)`, i.e.,
        `(top left, bottom right)`.
        """
        ret = self.tlwh.copy()
        ret[2:] += ret[:2]
        return ret

    def xyah(self):
        """Convert bounding box to format `(center x, center y, aspect ratio,
        height)`, where the aspect ratio is `width / height`.
        """
        ret = self.tlwh.copy()
        ret[:2] += ret[2:] / 2
        ret[2] /= ret[3]
        return ret

    @property
    def sly_label(self):
        return self._sly_label

    @sly_label.setter
    def sly_label(self, sly_label: Label):
        self._sly_label = sly_label

    def get_sly_label(self):
        return self.sly_label

    def set_sly_label(self, sly_label: Label):
        self.sly_label = sly_label

    def clean_sly_label(self):
        self.sly_label = None


class BaseTrack:
    def __init__(self, track_id, *args, **kwargs):
        self.track_id = track_id

    def get_sly_label(self):
        raise NotImplementedError()


class BaseTracker:
    def __init__(self, settings: Dict):
        self.settings = settings
        self.args = self.parse_settings(settings)
        self.device = self.select_device(device=self.args.device)

    def select_device(self, device="", batch_size=None):
        import torch  # pylint: disable=import-error

        # device = 'cpu' or '0' or '0,1,2,3'
        cpu_request = device.lower() == "cpu"
        if device and not cpu_request:  # if device requested other than 'cpu'
            os.environ["CUDA_VISIBLE_DEVICES"] = device  # set environment variable
            assert (
                torch.cuda.is_available()
            ), f"CUDA unavailable, invalid device {device} requested"  # check availablity

        cuda = False if cpu_request else torch.cuda.is_available()
        if cuda:
            c = 1024**2  # bytes to MB
            ng = torch.cuda.device_count()
            if ng > 1 and batch_size:  # check that batch_size is compatible with device_count
                assert (
                    batch_size % ng == 0
                ), f"batch-size {batch_size} not multiple of GPU count {ng}"
            x = [torch.cuda.get_device_properties(i) for i in range(ng)]
            s = f"Using torch {torch.__version__} "
            for i, d in enumerate((device or "0").split(",")):
                if i == 1:
                    s = " " * len(s)
                logger.info(f"{s}CUDA:{d} ({x[i].name}, {x[i].total_memory / c}MB)")
        else:
            logger.info(f"Using torch {torch.__version__} CPU")

        logger.info("")  # skip a line
        return torch.device("cuda:0" if cuda else "cpu")

    def parse_settings(self, settings: Dict) -> argparse.Namespace:
        _settings = self.default_settings()
        _settings.update(settings)
        if "device" not in _settings:
            _settings["device"] = ""
        return argparse.Namespace(**_settings)

    def default_settings(self):
        """To be overridden by subclasses."""
        return {}

    @contextmanager
    def _video_frames_generator(self, video_path: str):
        cap = cv2.VideoCapture(video_path)
        try:
            while cap.isOpened():
                ret, frame = cap.read()
                if not ret:
                    break
                yield frame
        finally:
            cap.release()

    def frames_generator(self, source: str):
        if isinstance(source, str):

            def _gen():
                with self._video_frames_generator(source) as frames:
                    for frame in frames:
                        yield frame

            return _gen()
        elif isinstance(source, list) and isinstance(source[0], str):
            return [cv2.imread(img) for img in source]
        else:
            return source

    def track(
        self,
        source: Union[List[np.ndarray], List[str], str],
        frame_to_annotation: Dict[int, Annotation],
        pbar_cb=None,
    ):
        """To be overridden by subclasses."""
        raise NotImplementedError()

    def convert_annotation(self, annotation_for_frame: Annotation):
        formatted_predictions = []
        sly_labels = []

        for label in annotation_for_frame.labels:
            confidence = 1.0
            if label.tags.get("confidence", None) is not None:
                confidence = label.tags.get("confidence").value
            elif label.tags.get("conf", None) is not None:
                confidence = label.tags.get("conf").value

            rectangle: Rectangle = label.geometry.to_bbox()
            tlwh = [
                rectangle.top,
                rectangle.left,
                rectangle.height,
                rectangle.width,
                confidence,
            ]

            formatted_predictions.append(tlwh)
            sly_labels.append(label)

        return formatted_predictions, sly_labels

    def update(
        self, img, annotation: Annotation, frame_index, tracks_data: Dict[int, List[Dict]] = None
    ):
        raise NotImplementedError()

    def correct_figure(self, img_size, figure):  # img_size — height, width tuple
        # check figure is within image bounds
        canvas_rect = Rectangle.from_size(img_size)
        if canvas_rect.contains(figure.to_bbox()) is False:
            # crop figure
            figures_after_crop = figure.crop(canvas_rect)
            if len(figures_after_crop) > 0:
                return figures_after_crop[0]
            else:
                return None
        else:
            return figure

    def update_track_data(self, tracks_data: dict, tracks: List[BaseTrack], frame_index: int):
        track_id_data = []
        labels_data = []

        for curr_track in tracks:
            track_id = curr_track.track_id

            if curr_track.get_sly_label() is not None:
                track_id_data.append(track_id)
                labels_data.append(curr_track.get_sly_label())

        tracks_data[frame_index] = {"ids": track_id_data, "labels": labels_data}

        return tracks_data

    def get_annotation(self, tracks_data: Dict, frame_shape, frames_count) -> VideoAnnotation:
        # Create and count object classes for each track
        object_classes = {}  # object_class_name -> object_class
        object_class_counter = {}  # track_id -> object_class_name -> count
        for frame_index, data in tracks_data.items():
            for track_id, label in zip(data["ids"], data["labels"]):
                label: Label
                object_classes.setdefault(label.obj_class.name, label.obj_class)
                object_class_counter.setdefault(track_id, {}).setdefault(label.obj_class.name, 0)
                object_class_counter[track_id][label.obj_class.name] += 1

        # Assign object classes to tracks
        track_obj_classes = {}  # track_id -> object_class
        for track_id, counters in object_class_counter.items():
            max_counter = -1
            obj_class_name = None
            for obj_class_name, count in counters.items():
                if count > max_counter:
                    max_counter = count
                    obj_class_name = obj_class_name
            track_obj_classes[track_id] = object_classes[obj_class_name]

        # Create video objects, figures and frames
        video_objects = {}  # track_id -> VideoObject
        frames = []
        for frame_index, data in tracks_data.items():
            frame_figures = []
            for track_id, label in zip(data["ids"], data["labels"]):
                label: Label
                video_object = video_objects.setdefault(
                    track_id, VideoObject(track_obj_classes[track_id])
                )
                frame_figures.append(VideoFigure(video_object, label.geometry, frame_index))
            frames.append(Frame(frame_index, frame_figures))

        objects = list(video_objects.values())
        return VideoAnnotation(
            img_size=frame_shape,
            frames_count=frames_count,
            objects=VideoObjectCollection(objects),
            frames=FrameCollection(frames),
        )
